#!/usr/bin/perl -wT
# This file is part of the webmpd front-end for MPD music player
#
# Copyright (C) 2006 Hannes Reich (hannes@skynet.ie)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

use strict;
use sigtrap;

use CGI;

use lib "."; # Change this to the full path of the dir containing the
             # script if MPD.pm is not found
use MPD;

use Encode ('encode');

# MPD connection details
use constant MPD_HOST => "localhost";
use constant MPD_PORT => "6600";

# Time intervals in seconds. Adjust for taste / bandwidth.
use constant REFRESH_INTERVAL => 300;

# How often to send an event, even if nothing is happening. Keep this
# below the watchdog timeout value in the javascript
use constant HEARTBEAT_INTERVAL => 5;

# How often to check mpd for status changes
use constant POLL_INTERVAL => 0.5;
# Use this to reduce burstiness of events
use constant MAX_EVENTS_PER_POLL => 500;

# TODO look into refreshing based on http volume sent
# TODO look into throttling

# explicit host and port below prevents taint warning that would
# result from allowing them to depend on environment variables.
my $mpd = new MPD(MPD_HOST, MPD_PORT);
my $cgi = new CGI;


sub get_status();
sub status_changed($$);
sub transmit_events($ );
sub quote($);
sub accurate_sleep($);

$| = 1; # autoflush output

print $cgi->header("text/html; encoding=".$cgi->charset()),
  $cgi->start_html(-script => 
                   'function u(key, newValue) {' . $cgi->param("callback") . '(key, newValue);}'
		   );

my @event_queue;

my %last_status;

my $next_refresh_time = time  + REFRESH_INTERVAL;
my $next_heartbeat_time = 0;
while(time < $next_refresh_time)
{
  my %status = get_status();
  for my $k (keys %status)
    {
      unshift @event_queue, status_changed($k, $status{$k})
	unless exists $last_status{$k} and $status{$k} eq $last_status{$k};
    }

  push @event_queue, status_changed("heartbeat", "")
    if(time > $next_heartbeat_time);
  $next_heartbeat_time = time + HEARTBEAT_INTERVAL
    if($#event_queue >= 0);

  transmit_events(\@event_queue);

  %last_status = %status;

  accurate_sleep(POLL_INTERVAL);
}

print $cgi->script("window.location.href='".$cgi->self_url()."';");
print $cgi->end_html;

exit;

# make a string safe to include in javascript
sub quote($)
  {
    ($_) = @_;
    return "" unless defined $_;
    s/\\/\\\\/g;
    s/"/\\"/g;
    return '"'.$_.'"';
  }

sub status_changed($$)
  {
    my($key, $value) = @_;
    return 'u('.quote($key).','.quote($value).");\n";
  }

sub transmit_events($)
  {
    my ($queue) = (@_);
    return
      unless @$queue;

    my $script = '';
    my $count = 0;
    while(@$queue && MAX_EVENTS_PER_POLL > $count)
      {
	$script .= shift @$queue;
	$count++;
      }
    print $cgi->script(encode($cgi->charset, $script));
  }


sub get_status()
  {
    $mpd->_get_status();
    my %status;
    map {
      $status{$_} = "";
      $status{$_} = $mpd->{$_}
	if defined $mpd->{$_};
    } ('repeat',
       'random',
       'state',
       'volume',
       'playlist',
       'playlistlength',
       'bitrate',
       'xfade',
       'audio',
       'song',
       'songid',
       'time',
       'updating_db'
      );

    my %song;
    %song = $mpd->get_song_info_from_id($status{'songid'})
      if $status{'songid'};
    map $status{'song.'.$_} = exists $song{$_} ? $song{$_} : "", (
								  'Artist',
								  'Title',
								  'Album',
								  'file'
								 );

    return %status;
  }

# sleep with subsecond accuracy (from perlfunc)
sub accurate_sleep($ )
  {
    select(undef, undef, undef, shift);
  }
